#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# pylint: disable=C0103

import logging
import re

log = logging.getLogger(__name__)

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    return [atoi(c) for c in re.split(r'(\d+)', text)]

def expand_range(interfaces):
    ''' Returns a naturally sorted list of items expanded from interfaces. '''

    match = re.match(r'^((([Ee](t(h)?)?(ernet)?)?(\d+)(([\/\-\,])?(\d+))*)'
                     '(,)?)+$', 
                     interfaces)

    if not match:
        raise TypeError('Unable to expand interface range: %s '
                        '(invalid input)' % interfaces)

    prefix = match.groups()[2]
    indicies_re = re.compile(r'[Ethernet|,]((\d(\/)?)+)(?!-)')
    range_re = re.compile(r'(\d+)\-(\d+)')

    indicies = [x[0] for x in indicies_re.findall(interfaces)]
    ranges = range_re.findall(interfaces)

    items = set()
    for start, end in ranges:
        start = int(start)
        end = int(end) + 1
        if end < start:
            log.warning('Unable to expand interface range: %s '
                        '(decreasing range)' % interfaces)
            raise TypeError('Unable to expand interface range: %s '
                            '(decreasing range)' % interfaces)            
        for index in range(start, end):
            items.add('%s%s' % (prefix, index))

    for index in indicies:
        if index == '0' or len(index.split('/')) > 3:
            log.warning('Unable to expand interface range: %s '
                        '(invalid index)' % interfaces)
            raise TypeError('Unable to expand interface range: %s '
                            '(invalid index)' % interfaces)            
        items.add('%s%s' % (prefix, index))

    items = list(items)
    items.sort(key=natural_keys)

    return items

def parse_interface(neighbor, node_id):
    try:
        if hasattr(neighbor, 'items'):
            remote_device = neighbor['device']
            remote_interface = neighbor.get('port', 'any')

        else:
            if neighbor == 'any':
                remote_device, remote_interface = 'any', 'any'
            elif neighbor == 'none':
                remote_device, remote_interface = 'none', 'none'
            elif ':' not in neighbor:
                remote_device = neighbor
                remote_interface = 'any'
            else:
                tokens = neighbor.split(':')
                remote_device = tokens[0]
                remote_interface = tokens[1]

        remote_device = str(remote_device).strip()
        if len(remote_device.split()) != 1:
            log.error('%s: interface parse error: invalid peer device %s' % 
                      (node_id, remote_device))
            raise Exception('%s: interface parse error: invalid peer '
                            'device %s' % (node_id, remote_device))
        
        remote_interface = str(remote_interface).strip()
        if len(remote_interface.split()) != 1:
            log.error('%s: interface parse error: invalid '
                      'peer interface %s' %
                      (node_id, remote_interface))
            raise Exception('%s: interface parse error: invalid peer '
                            'interface %s' % 
                            (node_id, remote_interface))

        return (remote_device, remote_interface)
    except KeyError as err:
        log.error('%s: interface parse error: missing attribute (%s)' % 
                  (node_id, str(err)))
        raise Exception('%s: interface parse error: missing '
                        'attribute (%s)' % (node_id, str(err)))



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# pylint: disable=W0622,W0402,W0613,W0142,R0201, E1103
#

import logging
import os
import routes
import urlparse

from string import Template
from webob.static import FileApp

import ztpserver.config

from ztpserver.constants import HTTP_STATUS_NOT_FOUND, HTTP_STATUS_CREATED
from ztpserver.constants import HTTP_STATUS_BAD_REQUEST, HTTP_STATUS_CONFLICT
from ztpserver.constants import HTTP_STATUS_INTERNAL_SERVER_ERROR
from ztpserver.constants import CONTENT_TYPE_JSON, CONTENT_TYPE_PYTHON
from ztpserver.constants import CONTENT_TYPE_YAML, CONTENT_TYPE_OTHER

from ztpserver.repository import create_repository
from ztpserver.repository import FileObjectNotFound, FileObjectError
from ztpserver.resources import ResourcePoolError
from ztpserver.serializers import SerializerError
from ztpserver.topology import create_node, load_pattern
from ztpserver.topology import load_neighbordb, resources
from ztpserver.topology import replace_config_action
from ztpserver.wsgiapp import WSGIController, WSGIRouter


DEFINITION_FN = 'definition'
STARTUP_CONFIG_FN = 'startup-config'
PATTERN_FN = 'pattern'
NODE_FN = '.node'
ATTRIBUTES_FN = 'attributes'
BOOTSTRAP_CONF = 'bootstrap.conf'

log = logging.getLogger(__name__)    # pylint: disable=C0103


class ValidationError(Exception):
    ''' Base exception class for :py:class:`Pattern` '''
    pass


class BaseController(WSGIController):

    FOLDER = None

    def __init__(self, **kwargs):
        data_root = ztpserver.config.runtime.default.data_root
        self.data_root = data_root
        self.repository = create_repository(data_root)
        super(BaseController, self).__init__()

    def expand(self, *args, **kwargs):
        ''' Returns an expanded file path relative to data_root '''

        file_path = os.path.join(*args)
        folder = kwargs.get('folder', self.FOLDER)
        return os.path.join(folder, file_path)

    def http_bad_request(self, *args, **kwargs):
        ''' Returns HTTP 400 Bad Request '''
        return dict(body='', content_type='text/html',
                    status=HTTP_STATUS_BAD_REQUEST)

    def http_not_found(self, *args, **kwargs):
        ''' Returns HTTP 404 Not Found '''

        return dict(body='', content_type='text/html',
                    status=HTTP_STATUS_NOT_FOUND)

    def http_internal_server_error(self, *args, **kwargs):
        ''' Returns HTTP 500 Internal server error '''

        return dict(body='', content_type='text/html',
                    status=HTTP_STATUS_INTERNAL_SERVER_ERROR)


class FilesController(BaseController):

    FOLDER = 'files'

    def __repr__(self):
        return 'FilesController(folder=%s)' % self.FOLDER

    def show(self, request, resource, **kwargs):
        ''' Handles GET /files/{resource} '''
        log.debug('%s\nResource: %s\n' % (request, resource))

        try:
            urlvars = request.urlvars
            if urlvars.get('format') is not None:
                resource += '.%s' % urlvars.get('format')
            file_path = self.expand(resource)
            filename = self.repository.get_file(file_path).name
            return FileApp(filename, content_type=CONTENT_TYPE_OTHER)
        except FileObjectNotFound:
            log.error('File %s not found' % resource)
            return self.http_not_found()


class ActionsController(BaseController):

    FOLDER = 'actions'

    def __repr__(self):
        return 'ActionsController(folder=%s)' % self.FOLDER

    def show(self, request, resource, **kwargs):
        ''' Handles GET /actions/{resource} '''
        log.debug('%s\nResource: %s\n' % (request, resource))

        try:
            file_path = self.expand(resource)
            body = self.repository.get_file(file_path).read(CONTENT_TYPE_PYTHON)
            return dict(body=body, content_type=CONTENT_TYPE_PYTHON)
        except FileObjectNotFound:
            log.error('Action %s not found' % resource)
            return self.http_not_found()


class NodesController(BaseController):

    FOLDER = 'nodes'

    def __repr__(self):
        return 'NodesController(folder=%s)' % self.FOLDER


    def fsm(self, state, **kwargs):
        ''' Execute the FSM for the request '''

        response = dict()
        try:
            while state != None:
                method = getattr(self, state)
                prev_state = state
                log.debug('%s: running %s' % (kwargs['node_id'], state))
                (response, state) = method(response, **kwargs)
        except ValidationError:            # pylint: disable=W0703
            log.error('%s: validation error in %s' % 
                      (kwargs['node_id'], prev_state))
            response = self.http_bad_request()
        except Exception as err:            # pylint: disable=W0703
            log.error('%s: error in %s: %s' % 
                      (kwargs['node_id'], prev_state, str(err)))
            response = self.http_bad_request()

        log.debug('%s: response to %s: %s' % 
                  (kwargs['node_id'], prev_state, response))
        return response                     # pylint: disable=W0150

    #-------------------------------------------------------------------

    def get_config(self, request, resource, **kwargs):
        log.debug('%s: node resource GET request: \n%s\n' % 
                  (resource, request))

        response = dict()

        try:
            filename = self.expand(resource, STARTUP_CONFIG_FN)
            response['body'] = self.repository.get_file(filename).read()
            response['content_type'] = CONTENT_TYPE_OTHER
        except FileObjectNotFound:
            log.error('%s: missing startup-config file %s' % 
                      (resource, filename))
            response = self.http_bad_request()
        except Exception as err:
            log.error('%s: unable to retrieve startup-config (%s)' %
                      (resource, err))
            response = self.http_bad_request()

        return response

    #-------------------------------------------------------------------

    def put_config(self, request, **kwargs):
        log.debug('%s: startup-config PUT request: \n%s\n' % 
                  (kwargs['resource'], request))

        response = dict()
        fobj = None
        try:
            body = str(request.body)
            content_type = str(request.content_type)
            filename = self.expand(kwargs['resource'], STARTUP_CONFIG_FN)
            fobj = self.repository.get_file(filename)
        except FileObjectNotFound:
            log.debug('%s: file not found: %s (adding it)' % 
                      (kwargs['resource'], filename))
            fobj = self.repository.add_file(filename)
        finally:
            if fobj:
                fobj.write(body, content_type)
            else:
                log.error('%s: unable to write %s' % 
                          (kwargs['resource'], filename))
                return self.http_bad_request()
        return response

    #-------------------------------------------------------------------

    def create(self, request, **kwargs):
        """ Handle the POST /nodes request

        The create method will handle in incoming POST request from the node
        and determine if the node already exists or not.  If the node
        does not exist, then the node will be created based on the
        request body.

        Args:
            request (webob.Request): the request object from WSGI

        Returns:
            A dict as the result of the state machine which is used to
            create a WSGI response object.

        """
        log.debug('%s\n' % request)

        try:
            node = create_node(request.json)
        except Exception as err:       # pylint: disable=W0703
            log.error('Unable to create node: %s (request=%s)' % (err, request))
            response = self.http_bad_request()
            return self.response(**response)

        node_id = node.identifier()
        if not node_id:
            log.error('Missing node identifier: %s (request=%s)' % 
                      (node, request))
            response = self.http_bad_request()
            return self.response(**response)

        return self.fsm('node_exists', request=request, 
                        node=node, node_id=node_id)

    def node_exists(self, response, *args, **kwargs):
        """ Checks if the node already exists and determines the next state

        This method will check for the existence of the node in the
        repository based on the node_id.  The node_id keyword is pulled
        from the kwargs dict.

        Args:
            response (dict): the response object being constructed
            kwargs (dict): arbitrary keyword arguments

        Returns:
            A tuple that includes the updated response object and the
            next state to transition to.  If the node already exists
            in the repository with a valid definition or startup-config,
            then the next state is 'dump_node' otherwise the next state
            is 'post_config'

        """
        next_state = 'post_config'
        node_id = kwargs.get('node_id')

        if self.repository.exists(self.expand(node_id, DEFINITION_FN)) or \
           self.repository.exists(self.expand(node_id, STARTUP_CONFIG_FN)):
            response['status'] = HTTP_STATUS_CONFLICT
            next_state = 'dump_node'
        else:
            if self.repository.exists(self.expand(node_id)):
                log.error('%s: node found on server, but no definition '
                          'or startup-config configured' % node_id)
                return (self.http_bad_request(), None)

        return (response, next_state)

    def post_config(self, response, *args, **kwargs):
        """ Writes the nodes startup config file if found in the request

        Args:
            response (dict): the response object being constructed
            kwargs (dict): arbitrary keyword arguments

        Returns:
            a tuple of response object and next state.  If a config key
            was found in the request, the next state is 'set_location'.
            If not, the next state is 'post_node'.

        """

        if 'config' not in kwargs['request'].json:
            # POST request for the node - will try to match neighbordb
            next_state = 'post_node'
        else:
            # POST request for the node's startup-config
            config = kwargs['request'].json['config']
            node_id = kwargs['node_id']

            self.repository.add_folder(self.expand(node_id))

            config_fn = self.expand(node_id, STARTUP_CONFIG_FN)
            self.repository.add_file(config_fn).write(config)

            response['status'] = HTTP_STATUS_CREATED
            next_state = 'set_location'

        return (response, next_state)

    def post_node(self, response, *args, **kwargs):
        """ Checks topology validation matches and writes node specific files

        This method will attempt to match the current node against the
        defined topology.  If a match is found, then the pattern matched
        and definition (defined in the pattern) are written to the nodes
        folder in the repository and the response status is set to HTTP
        201 Created.

        Args:
            response (dict): the response object being constructed
            kwargs (dict): arbitrary keyword arguments

        Returns:
            a tuple of response object and next state.  The next state
            is 'dump_node'

        Raises:
            If a match is not found, then a log message is created and
            an IndexError is raised.  If the node does not already
            exist in the repository, then a log message is created and a
            FileObjectNotFound exception is raised
            """

        node = kwargs['node']
        node_id = kwargs['node_id']
        
        neighbordb = load_neighbordb(node_id)
         # pylint: disable=E1103
        matches = neighbordb.match_node(node)
        if not matches:
            log.info('%s: node matched no patterns in neighbordb' %
                     node_id)
            return (self.http_bad_request(), None)

        log.debug('%s: %d pattern(s) in neihgbordb are a good match' %
                  (node_id, len(matches)))
        match = matches[0]

        log.info('%s: node matched \'%s\' pattern in neighbordb' % 
                 (node_id, match.name))
        try:

            definition_url = self.expand(match.definition, folder='definitions')
            fobj = self.repository.get_file(definition_url)
        except FileObjectNotFound:
            log.error('%s: failed to find definition (%s)' % 
                      (node_id, definition_url))
            raise

        try:
            definition = fobj.read(content_type=CONTENT_TYPE_YAML)
        except FileObjectError:
            log.error('%s: failed to load definition' % 
                      (node_id))
            raise

        definition_fn = self.expand(node_id, DEFINITION_FN)
        
        self.repository.add_folder(self.expand(node_id))

        log.info('%s: new /nodes/%s folder created' % 
                 (node_id, node_id))

        fobj = self.repository.add_file(definition_fn)
        fobj.write(definition, CONTENT_TYPE_YAML)
        
        pattern_fn = self.expand(node_id, PATTERN_FN)
        fobj = self.repository.add_file(pattern_fn)
        fobj.write(match.serialize(), CONTENT_TYPE_YAML)
        
        response['status'] = HTTP_STATUS_CREATED

        return (response, 'dump_node')

    def dump_node(self, response, *args, **kwargs):
        """ Writes the contents of the node to the repository

        Args:
            response (dict): the response object being constructed
            kwargs (dict): arbitrary keyword arguments

        Returns:
            a tuple of response object and next state.  The next state is
            'set_location'

        """

        try:
            node = kwargs.get('node')
            node_id = kwargs.get('node_id')
            contents = node.serialize()
            filename = self.expand(node_id, NODE_FN)
            fobj = self.repository.get_file(filename)
        except FileObjectNotFound:
            fobj = self.repository.add_file(filename)
        finally:
            fobj.write(contents, CONTENT_TYPE_JSON)
        return (response, 'set_location')

    def set_location(self, response, *args, **kwargs):
        """ Writes the HTTP Content-Location header

        Args:
            response (dict): the response object being constructed
            kwargs (dict): arbitrary keyword arguments

        Returns:
            a tuple of response object and next state.  The next state is
            None.

        Raises:
            Exception: catches a general exception for logging and then
                       re-raises it
        """
        node_id = kwargs.get('node_id')
        response['location'] = self.expand(node_id)
        return (response, None)

    #-------------------------------------------------------------------

    def show(self, request, resource, *args, **kwargs):
        """ Handle the GET /nodes/{resource} request

        Args:
            request (webob.Request): the request object from WSGI
            resource (str): the resource being requested

        Returns:
            A dict as the result of the state machine which is used to
            create a WSGI response object.

        """
        log.debug('%s\nResource: %s\n' % (request, resource))

        node_id = resource.split('/')[0]
        try:
            fobj = self.repository.get_file(self.expand(resource, NODE_FN))
            node = create_node(fobj.read(CONTENT_TYPE_JSON))
        except Exception as err:           # pylint: disable=W0703
            log.error('%s: unable to read file resource %s: %s' % 
                      (node_id, resource, err))
            response = self.http_bad_request()
            return self.response(**response)

        return self.fsm('get_definition', resource=resource, request=request,
                        node=node, node_id=node_id)

    def get_definition(self, response, *args, **kwargs):
        ''' Reads the node specific definition from disk and stores it in the
        repsonse dict with key `definition`
        '''

        try:
            filename = self.expand(kwargs['resource'], DEFINITION_FN)
            fobj = self.repository.get_file(filename)
            definition = fobj.read(CONTENT_TYPE_YAML,
                                   kwargs['resource'])
            response['definition'] = definition
            log.debug('%s: defintion is %s (%s)' % (kwargs['resource'], 
                                                    filename,
                                                    definition['actions']))
        except FileObjectNotFound:
            log.warning('%s: missing definition %s' % 
                        (kwargs['resource'], filename))
        except FileObjectError as err:
            log.error(err.message)
            raise Exception('failed to load definition %s' % filename)
        return (response, 'do_validation')

    def do_validation(self, response, *args, **kwargs):
        config = ztpserver.config.runtime
        if not config.default.disable_topology_validation:
            filename = self.expand(kwargs['resource'], PATTERN_FN)
            fobj = self.repository.get_file(filename)

            try:
                pattern = load_pattern(fobj.name, node_id=kwargs['resource'])
            except SerializerError as err:
                log.error(err.message)
                raise Exception('failed to load pattern %s' % filename)

            if not pattern:
                raise Exception('failed to validate pattern')

            if not pattern.match_node(kwargs['node']):
                log.error('%s: node failed pattern validation (%s)' % 
                          (kwargs['resource'], filename))
                raise ValidationError('%s: node failed pattern '
                                      'validation (%s)' %
                                      (kwargs['resource'], filename))
            log.debug('%s: node passed pattern validation (%s)' % 
                      (kwargs['resource'], filename))
        else:
            log.warning('%s: topology validation is disabled' %
                        kwargs['resource'])
        return (response, 'get_startup_config')

    def get_startup_config(self, response, *args, **kwargs):
        response['get_startup_config'] = False
        try:
            filename = self.expand(kwargs['resource'], STARTUP_CONFIG_FN)
            self.repository.get_file(filename)
            response['get_startup_config'] = True
            if 'definition' not in response:
                response['definition'] = dict(name='Autogenerated definition',
                                              actions=list())
            response['definition']['actions'].append(\
                replace_config_action(kwargs['resource'],
                                                           STARTUP_CONFIG_FN))
        except FileObjectNotFound:
            log.debug('%s: no startup-config %s' % 
                      (kwargs['resource'], filename))

        return (response, 'do_actions')

    def do_actions(self, response, *args, **kwargs):
        actions = response['definition']['actions']
        _actions = list()
        for action in actions:
            always_execute = action.get('always_execute', False)
            if always_execute:
                _actions.append(action)
                log.debug('%s: always_execute action %s included '
                          'in definition' %
                          (kwargs['resource'],  action.get('name')))
            elif not response['get_startup_config']:
                _actions.append(action)
                log.debug('%s: action %s included '
                          'in definition' %
                          (kwargs['resource'],  action.get('name')))
            else:
                log.debug('%s: action %s not included '
                          'in definition' %
                          (kwargs['resource'],  action.get('name')))
        response['definition']['actions'] = _actions
            
        return (response, 'get_attributes')

    def get_attributes(self, response, *args, **kwargs):
        ''' Reads the resource specific attributes file and stores it in the
        response dict as 'attributes'
        '''
        try:
            filename = self.expand(kwargs['resource'], ATTRIBUTES_FN)
            fileobj = self.repository.get_file(filename)
            attributes = fileobj.read(CONTENT_TYPE_YAML)
            response['attributes'] = attributes
            log.debug('%s: loaded %s attributes from %s' % 
                      (kwargs['resource'], attributes, filename))
        except FileObjectNotFound:
            log.warning('%s: no node specific attributes file' %
                        kwargs['resource'])
            response['attributes'] = dict()

        return (response, 'do_substitution')

    def do_substitution(self, response, *args, **kwargs):
        # pylint: disable=R0914
        definition = response.get('definition')
        attrs = definition.get('attributes', dict())

        nodeattrs = response.get('attributes', dict())

        def lookup(name):
            log.debug('%s: lookup up value for variable %s' % 
                      (kwargs['resource'], name))
            return nodeattrs.get(name, attrs.get(name))

        _actions = list()
        for action in definition['actions']:
            log.debug('%s: processing action %s (variable substitution)' %
                      (kwargs['resource'], action.get('name')))
            _attributes = dict()
            if 'attributes' in action:
                for key, value in action.get('attributes').items():
                    try:
                        update = dict()
                        for _key, _value in value.items():
                            if str(_value).startswith('$'):
                                _value = lookup(_value[1:])
                            update[_key] = _value
                    except AttributeError:
                        if str(value).startswith('$'):
                            value = lookup(value[1:])
                        update = value
                    finally:
                        _attributes[key] = update
            action['attributes'] = _attributes
            _actions.append(action)
        definition['actions'] = _actions
        response['definition'] = definition
        return (response, 'do_resources')

    def do_resources(self, response, *args, **kwargs):
        definition = response['definition']
        node = kwargs.get('node')
        _actions = list()

        try:
            for action in definition.get('actions'):
                attrs = action.get('attributes', dict())

                action['attributes'] = \
                    resources(attrs, node, kwargs['resource'])
                _actions.append(action)
        except ResourcePoolError as exc:
            log.error(exc)
            raise Exception('failed to allocate resources')
            
        definition['actions'] = _actions
        response['definition'] = definition
        return (response, 'finalize_response')

    def finalize_response(self, response, *args, **kwargs):
        _response = dict()
        _response['body'] = response['definition']
        _response['status'] = response.get('status', 200)
        _response['content_type'] = response.get('content_type',
                                                 CONTENT_TYPE_JSON)
        return (_response, None)



class BootstrapController(BaseController):

    DEFAULT_CONFIG = {
        'logging': list(),
        'xmpp': dict()
    }

    FOLDER = 'bootstrap'

    def __repr__(self):
        return 'BootstrapController(folder=%s)' % self.FOLDER

    def config(self, request, **kwargs):
        ''' Handles GET /bootstrap/config '''

        body = self.DEFAULT_CONFIG

        try:
            filename = self.expand(BOOTSTRAP_CONF)
            config = self.repository.get_file(filename).read(CONTENT_TYPE_YAML)
            if not config:
                log.warning('Bootstrap config file empty')                
            else:
                if 'logging' in config:
                    body['logging'] = config['logging']

                if 'xmpp' in config:
                    body['xmpp'] = config['xmpp'] 
                    for key in ['username', 'password', 'domain']:
                        if key not in body['xmpp']:
                            log.warning('Bootstrap config: \'%s\' missing from '
                                        'XMPP config' % key)
                    if 'rooms' not in body['xmpp'] or \
                       not body['xmpp']['rooms']:
                        log.warning('Bootstrap config: no XMPP rooms '
                                    'configured')
            resp = dict(body=body, content_type=CONTENT_TYPE_JSON)
        except FileObjectNotFound:
            log.warning('Bootstrap config file not found')
            resp = dict(body=body, content_type=CONTENT_TYPE_JSON)
        except FileObjectError:
            log.error('Failed to read bootstrap config file (%s)' % filename)
            resp = self.http_bad_request()
        except Exception as exc:
            log.error('Failed to load bootstrap config file (%s): %s' % 
                      (filename, exc))
            resp = self.http_bad_request()
        return resp

    def index(self, request, **kwargs):
        ''' Handles GET /bootstrap '''

        try:
            filename = self.expand(ztpserver.config.runtime.bootstrap.filename)
            fobj = self.repository.get_file(filename).read(CONTENT_TYPE_PYTHON)
            default_server = ztpserver.config.runtime.default.server_url
            body = Template(fobj).substitute(SERVER=default_server)
            resp = dict(body=body, content_type=CONTENT_TYPE_PYTHON)
        except KeyError as err:
            log.debug('Missing variable: %s' % err)
            resp = self.http_bad_request()
        except FileObjectNotFound:
            log.error('Bootstrap file not found (%s)' % filename)            
            resp = self.http_bad_request()
        except FileObjectError:
            log.error('Failed to read bootstrap file (%s)' % filename)
            resp = self.http_bad_request()
        return resp


class MetaController(BaseController):

    FOLDER = 'meta'

    BODY = {'size': None,
            'sha1': None}

    def __repr__(self):
        return 'MetaController(folder=%s)' % self.FOLDER

    def metadata(self, request, **kwargs):
        ''' Handles GET /meta/[actions|files|nodes]/<PATH_INFO> '''

        file_path = '%s/%s' % (kwargs['type'], kwargs['path_info'])

        try:
            try:
                file_resource = self.repository.get_file(file_path)
            except IOError as exc:
                # IOError is file_path points to a folder
                log.error('%s is a folder, not a file: %s' % 
                          (file_path, str(exc)))
                resp = self.http_not_found()
            else:
                self.BODY['size'] = file_resource.size()
                self.BODY['sha1'] = file_resource.hash()
                resp = dict(body=self.BODY, content_type=CONTENT_TYPE_JSON)
        except IOError as exc:
            log.error('Failed to collect meta information for %s: %s' %
                      (file_path, exc))
            resp = self.http_internal_server_error()
        return resp


class Router(WSGIRouter):
    ''' Routes incoming requests by mapping the URL to a controller '''

    def __init__(self):
        # pylint: disable=E1103,W0142

        mapper = routes.Mapper()

        kwargs = {}

        url = ztpserver.config.runtime.default.server_url
        log.debug('server URL: %s', url)
        parts = urlparse.urlsplit(url)
        if parts.path:
            path = parts.path[:-1] if parts.path.endswith('/') else parts.path
            if path:
                log.debug('path_prefix is %s', path)
                kwargs['path_prefix'] = path

        with mapper.submapper(**kwargs) as router_mapper:

            # configure /bootstrap
            router_mapper.connect('bootstrap', '/bootstrap',
                                  controller=BootstrapController,
                                  action='index',
                                  conditions=dict(method=['GET']))

            router_mapper.connect('bootstrap_config', '/bootstrap/config',
                                  controller=BootstrapController,
                                  action='config',
                                  conditions=dict(method=['GET']))


            # configure /meta
            router_mapper.connect('meta', 
                                  '/meta/{type:actions|files|nodes}/'
                                  '{path_info:.*}',
                                  controller=MetaController,
                                  action='metadata',
                                  conditions=dict(method=['GET']))

            # configure /nodes
            router_mapper.collection('nodes', 'node',
                                     controller=NodesController,
                                     collection_actions=['create'],
                                     member_actions=['show'],
                                     member_prefix='/{resource}')

            router_mapper.connect('get_node_config',
                                  '/nodes/{resource}/startup-config',
                                  controller=NodesController,
                                  action='get_config',
                                  conditions=dict(method=['GET']))

            router_mapper.connect('put_node_config',
                                  '/nodes/{resource}/startup-config',
                                  controller=NodesController,
                                  action='put_config',
                                  conditions=dict(method=['PUT']))

            # configure /actions
            router_mapper.collection('actions', 'action',
                                     controller=ActionsController,
                                     collection_actions=[],
                                     member_actions=['show'],
                                     member_prefix='/{resource}')

            # configure /files
            router_mapper.collection('files', 'file',
                                     controller=FilesController,
                                     collection_actions=[],
                                     member_actions=['show'],
                                     member_prefix='/{resource:.*}')

        super(Router, self).__init__(mapper)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=C0103,W0142
#
import collections
import logging
import os
import re
import string # pylint: disable=W0402
import ztpserver

import ztpserver.config

from ztpserver.validators import validate_neighbordb, validate_pattern
from ztpserver.resources import ResourcePool
from ztpserver.constants import CONTENT_TYPE_YAML
from ztpserver.serializers import load, SerializerError
from ztpserver.utils import expand_range, parse_interface

ANY_DEVICE_PARSER_RE = re.compile(r':(?=[any])')
NONE_DEVICE_PARSER_RE = re.compile(r':(?=[none])')
FUNC_RE = re.compile(r'(?P<function>\w+)(?=\(\S+\))\([\'|\"]'
                     r'(?P<arg>.+?)[\'|\"]\)')

ALL_CHARS = set([chr(c) for c in range(256)])
NON_HEX_CHARS = ALL_CHARS - set(string.hexdigits)

log = logging.getLogger(__name__)


Neighbor = collections.namedtuple('Neighbor', ['device', 'interface'])

def default_filename():
    ''' Returns the path for neighbordb based on the conf file
    '''

    filepath = ztpserver.config.runtime.default.data_root
    filename = ztpserver.config.runtime.neighbordb.filename
    return os.path.join(filepath, filename)

def load_file(filename, content_type, node_id):
    ''' Returns the contents of a file specified by filename.

    The requred content_type argument is required and indicates the
    text serialization format the contents are stored in.

    If the serializer load function encounters errors, None is returned

    '''
    try:
        return load(filename, content_type, node_id)
    except SerializerError:
        log.error('%s: failed to load file: %s' % (node_id, filename))
        raise

def load_neighbordb(node_id, contents=None):
    try:
        if not contents:
            contents = load_file(default_filename(), CONTENT_TYPE_YAML,
                                 node_id)

        # neighbordb is empty
        if not contents:
            contents = dict()

        if not validate_neighbordb(contents, node_id):
            log.error('%s: failed to validate neighbordb' % node_id)
            return

        neighbordb = Neighbordb(node_id)

        if 'variables' in contents:
            neighbordb.add_variables(contents['variables'])

        if 'patterns' in contents:
            neighbordb.add_patterns(contents['patterns'])

        log.debug('%s: loaded neighbordb: %s' % (node_id, neighbordb))
        return neighbordb
    except (NeighbordbError, SerializerError):
        log.error('%s: failed to load neighbordb' % node_id)
    except Exception as err:
        log.error('%s: failed to load neighbordb because of error: %s' %
                  (node_id, err))

def load_pattern(pattern, content_type=CONTENT_TYPE_YAML, node_id=None):
    """ Returns an instance of Pattern """
    try:
        if not isinstance(pattern, collections.Mapping):
            pattern = load_file(pattern, content_type,
                                node_id)

        if not validate_pattern(pattern, node_id):
            log.error('%s: failed to validate pattern attributes' % node_id)
            return None

        pattern['node_id'] = node_id
        return Pattern(**pattern)
    except TypeError:
        log.error('%s: failed to load pattern \'%s\'' % (node_id, pattern))

def create_node(nodeattrs):
    try:
        if nodeattrs.get('systemmac') is not None:
            _systemmac = nodeattrs['systemmac']
            for symbol in [':', '.']:
                _systemmac = str(_systemmac).replace(symbol, '')
            nodeattrs['systemmac'] = _systemmac
        node = Node(**nodeattrs)
        log.debug('%s: created node object %r' % (node.identifier(), node))
        return node
    except KeyError as err:
        log.error('Failed to create node - missing attribute: %s' % err)

def resources(attributes, node, node_id):
    log.debug('%s: computing resources (attr=%s)' % 
              (node_id, attributes))

    _attributes = dict()
    _resources = ResourcePool(node_id)

    for key, value in attributes.items():
        if hasattr(value, 'items'):
            value = resources(value, node, node_id)
        elif hasattr(value, '__iter__'):
            _value = list()
            for item in value:
                match = FUNC_RE.match(item)
                if match:
                    method = getattr(_resources, match.group('function'))
                    _value.append(method(match.group('arg'), node))
                else:
                    _value.append(item)
            value = _value
        else:
            match = FUNC_RE.match(str(value))
            if match:
                method = getattr(_resources, match.group('function'))
                value = method(match.group('arg'), node)
        _attributes[key] = value
    log.debug('%s: resources: %s' % (node_id, _attributes))
    return _attributes

def replace_config_action(resource, filename=None):
    ''' Builds a definition with a single action replace_config '''

    filename = filename or 'startup-config'
    server_url = ztpserver.config.runtime.default.server_url
    url = '%s/nodes/%s/%s' % (server_url, str(resource), filename)

    action = dict(name='install static startup-config file',
                  action='replace_config',
                  always_execute=True,
                  attributes={'url': url})

    return action

class NodeError(Exception):
    ''' Base exception class for :py:class:`Node` '''
    pass


class PatternError(Exception):
    ''' Base exception class for :py:class:`Pattern` '''
    pass


class InterfacePatternError(Exception):
    ''' Base exception class for :py:class:`InterfacePattern` '''
    pass


class NeighbordbError(Exception):
    ''' Base exception class for :py:class:`Neighbordb` '''
    pass


class OrderedCollection(collections.OrderedDict):
    ''' base object for using an ordered dictionary '''
    def __call__(self, key=None):
        #pylint: disable=W0221
        return self.get(key) if key else self.keys()


class Function(object):
    def __init__(self, value):
        self.value = value

    def match(self, arg):
        raise NotImplementedError


class IncludesFunction(Function):
    def match(self, arg):
        return self.value in arg


class ExcludesFunction(Function):
    def match(self, arg):
        return self.value not in arg


class RegexFunction(Function):
    def match(self, arg):
        match = re.match(self.value, arg)
        return match is not None


class ExactFunction(Function):
    def match(self, arg):
        return arg == self.value


class Node(object):
    ''' A Node object is maps the metadata from an EOS node.  It provides
    access to the node's meta data including interfaces and the
    associated neighbors found on those interfaces.
    '''

    def __init__(self, **kwargs):
        self.systemmac = kwargs.get('systemmac')
        self.model = kwargs.get('model')
        self.serialnumber = kwargs.get('serialnumber')
        self.version = kwargs.get('version')

        self.neighbors = OrderedCollection()
        if 'neighbors' in kwargs:
            self.add_neighbors(kwargs['neighbors'])

    def __repr__(self):
        return 'Node(serialnumber=%s, systemmac=%s, neighbors=%s)' % \
               (self.serialnumber, self.systemmac, self.neighbors)

    def identifier(self):
        identifier = ztpserver.config.runtime.default.identifier
        return getattr(self, identifier)

    def add_neighbor(self, interface, peers):
        try:
            if self.neighbors.get(interface):
                raise NodeError('%s: interface \'%s\' already added to node' % 
                                (self.identifier(), interface))

            _neighbors = list()
            for peer in peers:
                log.debug('%s: creating neighbor %s:%s for interface %s' %
                          ( self.identifier(), peer['device'], 
                            peer['port'], interface))
                _neighbors.append(Neighbor(peer['device'],
                                           peer['port']))
            self.neighbors[interface] = _neighbors
        except KeyError as err:
            log.error('%s: failed to neighbor because of missing key (%s)' % 
                      (self.identifier(), str(err)))
            raise NodeError('%s: failed to neighbor because of KeyError (%s)' % 
                      (self.identifier(), str(err)))

    def add_neighbors(self, neighbors):
        for interface, peers in neighbors.items():
            self.add_neighbor(interface, peers)

    def serialize(self):
        result = {}
        for prop in ['model', 'systemmac', 'serialnumber', 'version']:
            if getattr(self, prop):
                result[prop] = getattr(self, prop)

        neighbors = {}
        if self.neighbors:
            for interface, neighbor_list in self.neighbors.items():
                serialized_neighbor_list = []
                for neighbor in neighbor_list:
                    serialized_neighbor_list.append(
                        dict(device=neighbor.device, port=neighbor.interface))
                neighbors[interface] = serialized_neighbor_list
        result['neighbors'] = neighbors
        return result


class Neighbordb(object):

    RESERVED_VARIABLES = ['any', 'none']

    def __init__(self, node_id):
        self.node_id = node_id
        
        self.variables = dict()
        self.patterns = {'globals': list(), 'nodes': dict()}

    def __repr__(self):
        return 'Neighbordb(variables=%d, globals=%d, nodes=%d)' % \
               (len(self.variables),
                len(self.patterns['globals']),
                len(self.patterns['nodes']))

    def add_variable(self, key, value, overwrite=False):
        if key in self.RESERVED_VARIABLES:
            log.error('%s: failed to add variable: %s (reserved keyword)' % 
                      (self.node_id, key))
            raise NeighbordbError('%s: failed to add variable: %s '
                                '(reserved keyword)' % (self.node_id, key))
        elif key in self.variables and not overwrite:
            log.error('%s: failed to add variable: %s (duplicate)' % 
                      (self.node_id, key))
            raise NeighbordbError('%s: failed to add variable %s '
                                '(duplicate)' % (self.node_id, key))

        self.variables[key] = value

    def add_variables(self, variables):
        if not hasattr(variables, 'items'):
            log.error('%s: failed to add variables: missing attribute '
                      '\'items\' (%s)' % (self.node_id, variables))
            raise NeighbordbError('%s: failed to add variables: '
                                  'missing attribute \'items\' (%s)' % 
                                  (self.node_id, variables))

        for key, value in variables.items():
            self.add_variable(key, value)

    def add_pattern(self, name, **kwargs):

        try:
            kwargs['node_id'] = self.node_id
            kwargs['name'] = name

            kwargs['node'] = kwargs.get('node')
            kwargs['definition'] = kwargs.get('definition')
            kwargs['interfaces'] = kwargs.get('interfaces', list())
            kwargs['variables'] = kwargs.get('variables', dict())

            for key in set(self.variables).difference(kwargs['variables']):
                kwargs['variables'][key] = self.variables[key]

                
            pattern = Pattern(**kwargs)

            log.debug('%s: pattern \'%r\' parsed successfully' % 
                      (self.node_id, pattern))

            # Add pattern to neighbordb
            if kwargs['node']:
                self.patterns['nodes'][pattern.node] = pattern
            else:
                self.patterns['globals'].append(pattern)
        except KeyError as err:
            log.error('%s: failed to add pattern \'%s\' because of '
                      'missing key (%s)' % (self.node_id, name, str(err)))
            raise NeighbordbError('%s: failed to pattern \'%s\' because of '
                                'missing key (%s)' % 
                                  (self.node_id, name, str(err)))
        except PatternError:
            log.error('%s: failed to add pattern \'%s\'' % 
                      (self.node_id, name))
            raise NeighbordbError('%s: failed to add pattern \'%s\'' % 
                                  (self.node_id, name))

    def add_patterns(self, patterns):
        try:
            for pattern in patterns:
                self.add_pattern(**pattern)
        except TypeError as err:
            log.error('%s: failed to add patterns %s: %s' % 
                      (self.node_id, patterns, str(err)))
            raise NeighbordbError('%s: failed to add patterns %s: %s' %
                                  (self.node_id, patterns, str(err)))

    def is_node_pattern(self, pattern):
        #pylint: disable=R0201
        return pattern.node

    def is_global_pattern(self, pattern):
        #pylint: disable=R0201
        return not pattern.node

    def get_patterns(self):
        return self.patterns['nodes'].values() + self.patterns['globals']

    @staticmethod
    def identifier(node):
        identifier = ztpserver.config.runtime.default.identifier
        return node[identifier]
        
    def find_patterns(self, node):
        identifier = node.identifier()
        log.debug('%s: searching for eligible patterns' % 
                  identifier)
        pattern = self.patterns['nodes'].get(identifier, None)
        if pattern:
            log.debug('%s: eligible pattern: %s' % (identifier, 
                                                    pattern.name))
            return [pattern]
        else:
            if self.patterns['globals']:
                log.debug('%s: all global patterns are eligible' %
                          identifier)
            else:
                log.debug('%s: no eligible patterns' %
                          identifier)
            return self.patterns['globals']

    def match_node(self, node):
        identifier = node.identifier()
        result = list()
        for pattern in self.find_patterns(node):
            log.debug('%s: attempting to match pattern %s' % 
                      (identifier, pattern.name))
            if pattern.match_node(node):
                log.debug('%s: pattern %s matched' % 
                          (identifier, pattern.name))
                result.append(pattern)
            else:
                log.debug('%s: pattern %s match failed' % 
                          (identifier, pattern.name))
        return result


class Pattern(object):

    def __init__(self, name=None, definition=None, interfaces=None,
                 node=None, variables=None, node_id=None):

        self.name = name
        self.definition = definition

        self.node = node
        self.node_id = node_id
        self.variables = variables or dict()

        self.interfaces = list()
        if interfaces:
            self.add_interfaces(interfaces)

        self.variable_substitution()

    def __repr__(self):
        return 'Pattern(name=\'%s\')' % self.name

    def variable_substitution(self):
        try:
            log.debug('%s: checking pattern \'%s\' entries for variable '
                      'substitution' % (self.node_id, self.name))
            for entry in self.interfaces:
                for item in entry['patterns']:
                    for attr in ['remote_device', 'remote_interface']:
                        value = getattr(item, attr)
                        if value.startswith('$'):
                            newvalue = self.variables[value[1:]]
                            setattr(item, attr, newvalue)
                    item.refresh()
            log.debug('%s: pattern \'%s\' variable substitution complete' %
                      (self.node_id, self.name))
        except KeyError as exc:
            log.debug('%s: pattern \'%s\' variable substitution failed: %s' %
                      (self.node_id, self.name, str(exc)))
            raise PatternError('%s: pattern \'%s\' variable substitution '
                               'failed: %s' % 
                               (self.node_id, self.name, str(exc)))

    def serialize(self):
        data = dict(name=self.name, definition=self.definition)
        data['variables'] = self.variables
        data['node'] = self.node

        interfaces = []
        for item in self.interfaces:
            _item = item['metadata']
            interfaces.append({_item['interface']: _item['neighbors']})
        data['interfaces'] = interfaces

        return data

    def parse_interface(self, neighbor):
        try:
            return parse_interface(neighbor, self.node_id)
        except Exception as err:
            raise PatternError(str(err))

    def add_interface(self, interface):
        try:
            if not hasattr(interface, 'items'):
                log.error('%s: pattern \'%s\' - failed to add interface %s: '
                          'missing attribute (items)' %
                          (self.node_id, self.name, interface))
                raise PatternError('%s: pattern \'%s\' - failed to add '
                                   'interface %s: missing attribute (items)' %
                                   (self.node_id, self.name, interface))

            for intf, neighbors in interface.items():
                (remote_device, remote_interface) = \
                    self.parse_interface(neighbors)

                metadata = dict(interface=intf, neighbors=neighbors)

                patterns = list()
                if intf in ['none', 'any']:
                    patterns.append(InterfacePattern(intf, remote_device, 
                                                     remote_interface, 
                                                     self.node_id))
                else:
                    for item in expand_range(intf):
                        pattern = InterfacePattern(item, remote_device, 
                                                   remote_interface,
                                                   self.node_id)
                        patterns.append(pattern)
                self.interfaces.append(dict(metadata=metadata,
                                            patterns=patterns))
        except InterfacePatternError:
            log.error('%s: pattern \'%s\' - failed to add interface %s' %
                      (self.node_id, self.name, interface))
            raise PatternError('%s: pattern \'%s\' - failed to add '
                               'interface %s' %
                               (self.node_id, self.name, interface))

    def add_interfaces(self, interfaces):
        try:
            for interface in interfaces:
                self.add_interface(interface)
        except TypeError as err:
            log.error('%s: pattern \'%s\' - failed to add interfaces %s: %s' %
                      (self.node_id, self.name, interface, str(err)))
            raise PatternError('%s: pattern \'%s\' - failed to add '
                               'interfaces %s: %s' %
                               (self.node_id, self.name, interface, str(err)))

    def match_node(self, node):

        log.debug('%s: pattern \'%s\' - attempting to match node (%r)' % 
                  (self.node_id, self.name, str(node)))

        patterns = list()
        for entry in self.interfaces:
            for pattern in entry['patterns']:
                patterns.append(pattern)

        for interface, neighbors in node.neighbors.items():
            log.debug('%s: pattern \'%s\' - attempting to match '
                      'interface %s(%s)' %
                      (self.node_id, self.name, interface, str(neighbors)))

            match = False
            for index, pattern in enumerate(patterns):
                log.debug('%s: pattern \'%s\' - checking interface pattern '
                          'for %s: %s' % 
                          (self.node_id, self.name, interface, pattern))
                result = pattern.match(interface, neighbors)

                # True, False, None
                if result is True:
                    log.debug('%s: pattern \'%s\' - interface pattern match '
                              'for %s: %s' % 
                              (self.node_id, self.name, interface, pattern))
                    del patterns[index]
                    match = True
                    break
                elif result is False:
                    log.debug('%s: pattern \'%s\' - interface pattern match '
                              'failure for %s: %s' % 
                              (self.node_id, self.name, interface, pattern))
                    return False

            if not match:
                log.debug('%s: pattern \'%s\' - interface %s did not match '
                          'any interface patterns' % 
                          (self.node_id, self.name, interface))

        for pattern in patterns:
            if pattern.is_positive_constraint():
                log.debug('%s: pattern \'%s\' - interface pattern %s did '
                          'not match any interface' % 
                          (self.node_id, self.name, pattern))
                return False
        return True


class InterfacePattern(object):

    KEYWORDS = {
        'any': RegexFunction('.*'),
        'none': RegexFunction('[^a-zA-Z0-9]')
    }

    FUNCTIONS = {
        'exact': ExactFunction,
        'includes': IncludesFunction,
        'excludes': ExcludesFunction,
        'regex': RegexFunction
    }

    def __init__(self, interface, remote_device, remote_interface, node_id):

        match = re.match(r'^[ehnrtE]+(\d.*)$', interface)
        if match:
            self.interface = 'Ethernet%s' % match.groups()[0]
        else:
            self.interface = interface

        self.remote_device = remote_device
        self.remote_interface = remote_interface
        self.node_id = node_id

        self.remote_device_re = self.compile(remote_device)
        self.remote_interface_re = self.compile(remote_interface)

    def __repr__(self):
        return 'InterfacePattern(interface=%s, remote_device=%s, ' \
               'remote_interface=%s)' % \
                (self.interface, self.remote_device, self.remote_interface)

    def refresh(self):
        self.remote_device_re = self.compile(self.remote_device)
        self.remote_interface_re = self.compile(self.remote_interface)

    def compile(self, value):
        if value in self.KEYWORDS:
            return self.KEYWORDS[value]

        try:
            match = FUNC_RE.match(value)
            if match:
                function = match.group('function')
                arg = match.group('arg')
                return self.FUNCTIONS[function](arg)
            else:
                return ExactFunction(value)
        except KeyError as exc:
            log.error('%s: compile error: unknown function \'%s\' (%s)' % 
                      (self.node_id, function, str(exc)))
            raise InterfacePatternError

    def match(self, interface, neighbors):
        for neighbor in neighbors:
            res = self.match_neighbor(interface, Neighbor(neighbor.device,
                                                          neighbor.interface))
            if res is True:
                return True
            elif res is False:
                return False
        return None

    def match_neighbor(self, interface, neighbor):
        # pylint: disable=R0911,R0912
        log.debug('%s: attempting to match %s(%s) against '
                  'interface pattern %r' % 
                  (self.node_id, interface, neighbor, self))
        
        if self.interface == 'any':
            if self.remote_device == 'any':
                if self.remote_interface == 'any':
                    return True
                elif self.remote_interface == 'none':
                    # bogus
                    return False
                else:
                    if self.match_remote_interface(neighbor.interface):
                        return True
            elif self.remote_device == 'none':
                if self.remote_interface == 'any':
                    # bogus
                    return False
                elif self.remote_interface == 'none':
                    # bogus
                    return False
                else:
                    return False
            else:
                if self.remote_interface == 'any':
                    if self.match_remote_device(neighbor.device):
                        return True
                elif self.remote_interface == 'none':
                    if self.match_remote_device(neighbor.device):
                        return False 
                else:
                    if(self.match_remote_device(neighbor.device) and
                       self.match_remote_interface(neighbor.interface)):
                        return True

        elif self.interface == 'none':
            if self.remote_device == 'any':
                if self.remote_interface == 'any':
                    # bogus
                    return False
                elif self.remote_interface == 'none':
                    # bogus
                    return False
                else:
                    if self.match_remote_interface(neighbor.interface):
                        return False
            elif self.remote_device == 'none':
                if self.remote_interface == 'any':
                    # bogus
                    return False
                elif self.remote_interface == 'none':
                    # no LLDP capable neighbors
                    return False
                else:
                    # bogus
                    return False
            else:
                if self.remote_interface == 'any':
                    if self.match_remote_device(neighbor.device):
                        return False
                elif self.remote_interface == 'none':
                    if self.match_remote_device(neighbor.device):
                        return False
                else:
                    if(self.match_remote_device(neighbor.device) and
                       self.match_remote_interface(neighbor.interface)):
                        return False
        else:
            if self.remote_device == 'any':
                if self.remote_interface == 'any':
                    if self.match_interface(interface):
                        return True
                elif self.remote_interface == 'none':
                    if self.match_interface(interface):
                        return False
                else:
                    if(self.match_interface(interface) and
                        self.match_remote_interface(neighbor.interface)):
                        return True
            elif self.remote_device == 'none':
                if self.remote_interface == 'any':
                    if self.match_interface(interface):
                        return False
                elif self.remote_interface == 'none':
                    if self.match_interface(interface):
                        return False
                else:
                    if(self.match_interface(interface) and
                        self.match_remote_interface(neighbor.interface)):
                        return False
            elif self.match_interface(interface):
                if self.remote_interface == 'any':
                    if self.match_remote_device(neighbor.device):
                        return True
                elif self.remote_interface == 'none':
                    if self.match_remote_device(neighbor.device):
                        return False
                else:
                    if(self.match_interface(interface) and
                       self.match_remote_device(neighbor.device) and
                       self.match_remote_interface(neighbor.interface)):
                        return True 
        return None


    def match_interface(self, interface):
        if self.interface == 'any':
            return True
        elif self.interface is None:
            return False
        else:
            return self.interface == interface

    def match_remote_device(self, remote_device):
        if self.remote_device == 'any':
            return True
        elif self.remote_device is None:
            return False
        else:
            return self.remote_device_re.match(remote_device)

    def match_remote_interface(self, remote_interface):
        if self.interface == 'any':
            return True 
        elif self.interface is None:
            return False 
        else:
            return self.remote_interface_re.match(remote_interface)

    def is_positive_constraint(self):
        if self.interface == 'any':
            if self.remote_device == 'any':
                return True
            elif self.remote_device != 'none':
                return self.interface != 'none'

        elif self.interface != 'none':
            if self.remote_device == 'any':
                return True
            elif self.remote_device != 'none':
                return self.interface != 'none'



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=C0103
#
import os
import sys
import argparse

import logging

from wsgiref.simple_server import make_server

import ztpserver.config
import ztpserver.controller
import ztpserver.topology

from ztpserver.serializers import load
from ztpserver.validators import NeighbordbValidator
from ztpserver.constants import CONTENT_TYPE_YAML
from ztpserver.topology import default_filename

from ztpserver import __version__ as VERSION

DEFAULT_CONF = '/etc/ztpserver/ztpserver.conf'

log = logging.getLogger("ztpserver")
log.setLevel(logging.DEBUG)
log.addHandler(logging.NullHandler())

def enable_handler_console(level=None):
    """ Enables logging to stdout """
    
    logging_fmt = ztpserver.config.runtime.default.console_logging_format
    formatter = logging.Formatter(logging_fmt)

    ch = logging.StreamHandler()
    ch.tag = 'console'

    
    for handler in log.handlers:
        if 'tag' in handler.__dict__ and handler.tag == ch.tag:
            # Handler previously added
            return

    level = level or 'DEBUG'
    level = str(level).upper()
    level = logging.getLevelName(level)
    ch.setLevel(level)
    ch.setFormatter(formatter)

    log.addHandler(ch)

def python_supported():
    """ Returns True if the current version of the python runtime is valid """
    return sys.version_info > (2, 7) and sys.version_info < (3, 0)

def start_logging(debug):
    """ reads the runtime config and starts logging if enabled """

    if ztpserver.config.runtime.default.logging:
        if ztpserver.config.runtime.default.console_logging:
            enable_handler_console('DEBUG' if debug else 'INFO')

def load_config(conf=None):
    conf = conf or DEFAULT_CONF
    conf = os.environ.get('ZTPS_CONFIG', conf)

    if os.path.exists(conf):
        ztpserver.config.runtime.read(conf)

def start_wsgiapp(conf=None, debug=False):
    """ Provides the entry point into the application for wsgi compliant
    servers.   Accepts a single keyword argument ``conf``.   The ``conf``
    keyword argument specifies the path the server configuration file.  The
    default value is /etc/ztpserver/ztpserver.conf.

    :param conf: string path pointing to configuration file
    :return: a wsgi application object

    """

    load_config(conf)
    start_logging(debug)

    log.info('Logging started for ztpserver')
    log.info('Using repository %s', ztpserver.config.runtime.default.data_root)

    if not python_supported():
        raise SystemExit('ERROR: ZTPServer requires Python 2.7')

    return ztpserver.controller.Router()

def run_server(conf, debug):
    """ The :py:func:`run_server` is called by the main command line routine to
    run the server as standalone.   This function accepts a single argument
    that points towards the configuration file describing this server

    This function will block on the active thread until stopped.

    :param conf: string path pointing to configuration file
    """

    app = start_wsgiapp(conf, debug)

    host = ztpserver.config.runtime.server.interface
    port = ztpserver.config.runtime.server.port

    httpd = make_server(host, port, app)

    print "Starting server on http://%s:%s" % (host, port)

    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        print 'Shutdown'

def run_validator(filename=None):

    try:
        print 'Validating file \'%s\'\n' % filename
        validator = NeighbordbValidator()
        filename = filename or default_filename()
        validator.validate(load(filename, CONTENT_TYPE_YAML,
                                'validator'))
        print 'Valid Patterns (count: %d)' % len(validator.valid_patterns)
        print '--------------------------'
        for index, pattern in enumerate(sorted(validator.valid_patterns)):
            print '[%d] %s' % (index, pattern[1])
        print
        print 'Failed Patterns (count: %d)' % len(validator.invalid_patterns)
        print '---------------------------'
        for index, pattern in enumerate(sorted(validator.invalid_patterns)):
            print '[%d] %s' % (index, pattern[1])
        print

    except Exception as exc:        #pylint: disable=W0703
        log.exception(exc)
        print 'An unexpected error occurred trying to run the validator'



def main():
    """ The :py:func:`main` is the main entry point for the ztpserver if called
    from the commmand line.   When called from the command line, the server is
    running in standalone mode as opposed to using the :py:func:`application` to
    run under a python wsgi compliant server
    """

    usage = 'ztpserver [options]'

    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument('--version', '-v',
                        action='store_true',
                        help='Displays the version information')

    parser.add_argument('--conf', '-c',
                        type=str,
                        default=DEFAULT_CONF,
                        help='Specifies the configuration file to use')

    parser.add_argument('--validate',
                        type=str,
                        metavar='FILENAME',
                        help='Runs a validation check on neighbordb')

    parser.add_argument('--debug',
                        action='store_true',
                        help='Enables debug output to the STDOUT')


    args = parser.parse_args()
    if args.version:
        print 'ztps version %s' % VERSION
        sys.exit()

    if args.validate is not None:
        load_config(args.conf)
        start_logging(args.debug)
        sys.exit(run_validator(args.validate))

    return run_server(args.conf, args.debug)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import logging

import ztpserver.config

from ztpserver.serializers import load, dump
from ztpserver.constants import CONTENT_TYPE_YAML

log = logging.getLogger(__name__)   #pylint: disable=C0103


class ResourcePoolError(Exception):
    ''' base error raised by :py:class:`Resource` '''
    pass


class ResourcePool(object):

    def __init__(self, node_id):
        self.node_id = node_id

        cfg = ztpserver.config.runtime
        self.file_path = os.path.join(cfg.default.data_root, 'resources')
        self.data = None

    def serialize(self):
        data = dict()
        for key, value in self.data.items():
            data[key] = str(value) if value is not None else None
        return data

    def load(self, pool):
        self.data = dict()
        filename = os.path.join(self.file_path, pool)
        contents = load(filename, CONTENT_TYPE_YAML, self.node_id)
        for key, value in contents.items():
            self.data[key] = str(value) if value is not None else None

    def dump(self, pool):
        file_path = os.path.join(self.file_path, pool)
        dump(self, file_path, CONTENT_TYPE_YAML, self.node_id)

    def allocate(self, pool, node):
        node_id = node.identifier()
        log.debug('%s: allocating resources' % node_id)

        match = self.lookup(pool, node)

        try:
            if match:
                log.debug('%s: already allocated resource \'%s\':\'%s\'' % 
                         (node_id, pool, match))
                return match

            self.load(pool)
            key = next(x[0] for x in self.data.items() if x[1] is None)

            log.debug('%s: allocated \'%s\':\'%s\'' %
                      (node_id, pool, key))

            self.data[key] = node_id
            self.dump(pool)
        except StopIteration:
            log.error('%s: no resource free in \'%s\'' % (node_id, pool))
            raise ResourcePoolError('%s: no resource free in \'%s\'' % 
                                    (node_id, pool))
        except Exception as exc:
            log.error('%s: failed to allocate resource from \'%s\'' % 
                      (node_id, pool))
            raise ResourcePoolError(exc.message)
        return str(key)

    def lookup(self, pool, node):
        ''' Return an existing allocated resource if one exists '''
        node_id = node.identifier()
        log.debug('%s: looking up resource in \'%s\'' % 
                  (node_id, pool))
        try:
            self.load(pool)
            matches = [m[0] for m in self.data.iteritems()
                       if m[1] == node_id]
            key = matches[0] if matches else None
            return key
        except Exception as exc:
            log.error('%s: failed to lookup resource from \'%s\'' % 
                      (node_id, pool))
            raise ResourcePoolError(exc.message)



# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from glob import glob
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from ztpserver import __version__, __author__

PACKAGES = ['ztpserver']

INSTALL_ROOT = os.getenv('VIRTUAL_ENV', '')
if os.environ.get('READTHEDOCS'):
    print "Customizing install for ReadTheDocs.org build servers..."
    INSTALL_ROOT = "."
    from subprocess import call
    call(['docs/setup_rtd_files.sh'])
    PACKAGES.append('ztps')

CONF_PATH = INSTALL_ROOT + '/etc/ztpserver'
INSTALL_PATH = INSTALL_ROOT + '/usr/share/ztpserver'
INSTALL_REQUIREMENTS = open('requirements.txt').read().split('\n')

setup(
    name='ztpserver',
    version=__version__,
    description = 'ZTP Server for EOS',
    author=__author__,
    author_email='eosplus-dev@arista.com',
    url='https://github.com/arista-eosplus/ztpserver',
    download_url='https://github.com/arista-eosplus/ztpserver/tarball/v1.1.0',
    license='BSD-3',
    install_requires=INSTALL_REQUIREMENTS,
    packages=PACKAGES,
    scripts=glob('bin/*'),
    data_files=[
        ('%s/nodes' % INSTALL_PATH, []),
        ('%s/definitions' % INSTALL_PATH, []),
        ('%s/files' % INSTALL_PATH, []),
        ('%s/resources' % INSTALL_PATH, []),
        (CONF_PATH, glob('conf/ztpserver.conf')),
        ('%s/bootstrap' % INSTALL_PATH, glob('client/bootstrap')),
        ('%s/bootstrap' % INSTALL_PATH, glob('conf/bootstrap.conf')),
        ('%s/actions' % INSTALL_PATH, glob('actions/*')),
        ('%s' % INSTALL_PATH, glob('conf/neighbordb')),

        # 4.12.x support
        ('%s/files/lib' % INSTALL_PATH, glob('client/lib/requests-2.3.0.tar.gz')),
    ]
)
